// Joystick Shield with 2x16 lcd - Arnoud, 27th january 2021
// See https://www.whizzbizz.com/en/bluetooth.joystick.shield
// Coded to be 'ArduinoBlue' compatible. See https://sites.google.com/stonybrook.edu/arduinoble/

#include "LiquidCrystal_I2C.h"

// Defines to be ArduinoBlue compatible...
#define PATH_TRANSMISSION 244
#define PATH_TRANSMISSION_CONFIRMATION 245
#define LOCATION_TRANSMISSION_START 246
#define DELIMETER 247
#define TEXT_SEND_TRANSMISSION 248
#define CONNECTION_CHECK 249
#define TRANSMISSION_END 250
#define DRIVE_TRANSMISSION 251
#define BUTTON_TRANSMISSION 252
#define SLIDER_TRANSMISSION 253
#define TEXT_TRANSMISSION 254
#define DEFAULT_VALUE 255
#define TEXT_TRANSMISSION_TIMEOUT 5000 // ms
#define SHORT_TRANSMISSION_TIMEOUT 500
#define PATH_TRANSMISSION_TIMEOUT 10000
#define PATH_OVERFLOW_VALUE 1000000
// Some variables to be ArduinoBlue compatible...
const uint8_t DEFAULT_STEERING = 49;
const uint8_t DEFAULT_THROTTLE = 49;
const uint8_t MAX_SHORT_SIGNAL_LENGTH = 3;

#if defined(ARDUINO) && ARDUINO >= 100
#define printByte(args) write(args);
#else
#define printByte(args) print(args,BYTE);
#endif

#define button_1  2 // Button Blue - A
#define button_2  3 // Button Yellow - B
#define button_3  4 // Button Blue - C 
#define button_4  5 // Button Yellow - D 
#define button_5  7 // SMD button E on pcb
#define button_6  6 // SMD button F on pcb
#define button_7  8 // Button in joystick
#define x_axis A0   // X axis of joystick
#define y_axis A1   // Y axis of joystick
#define tone_pin A3 // Output to sound a tone
#define bt_state A2 // Bluetooth connection state read out from module

int buttons[]={ button_1, button_2, button_3, button_4, button_5, button_6, button_7 };
bool show_values = true; // Initializes with button/stick values on display, button E toggles manually...
int current_x_value;
int current_y_value;
int last_x_value;
int last_y_value;
int tone_freq = 0;
int tone_time = 0;

#define MAX_LINE 1 // LCD line is either 0 or 1
#define MAX_COL 15 // LCD column is 0 to 15 for 16x2 (1602)

int lcd_line;
int lcd_col;
String receive_buffer; // Values retreived from the other side...
String lcd_line0;  // Be sure to provide a Sting for each line
String lcd_line1;  // So we can rewrite lcd if blinking...
String empty_line; // Used to quickly CLR one line...
bool blink_line0 = false;
bool blink_line1 = false;
bool showing_line0 = true;
bool showing_line1 = true;
bool blink_show = true; // Toggling every 500ms
bool bt_connected = false;
unsigned long blinktime;

// BT logo was designed with https://maxpromer.github.io/LCD-Character-Creator/
uint8_t BTsymbol[8]  = {0x04,0x06,0x15,0x0E,0x04,0x0E,0x15,0x06};

//#define DEBUG // Outcomment for stealth/live operation...

// Setting of baud rate: set to 9600 to use default settings of most BT-modules
// Set your HC-05 module with AT+UART=38400,1,0 or HM-10/-19 module with AT+BAUD2
//#define BTBAUD 38400
#define BTBAUD 9600

LiquidCrystal_I2C lcd(0x27,16,2);  // set the LCD address to 0x27 for a 16 chars and 2 line display

void setup() {
  for (int i=0; i <7 ; i++) {
    pinMode(buttons[i], INPUT_PULLUP);
    digitalWrite(buttons[i], HIGH);  
  }
  pinMode(tone_pin, OUTPUT);
  pinMode(bt_state, INPUT);

  empty_line = "";
  for (int i=0; i<=MAX_COL; i++) {
    empty_line.concat((char)' ');
  }

  // Setup lcd display...
  lcd.init(); // initialize the lcd...
  lcd.backlight();
  lcd.setCursor(1,0);
  lcd.print("WHIZZBIZZ.COM");
  lcd.setCursor(1,1);
  lcd.print("Remote Control");
  lcd.createChar(0, BTsymbol);
  delay(2000);
  lcd.clear();
  
  // Setup serial to be used for BT...
  Serial.begin(BTBAUD); 
}

// slaveReadString reads value message from the BT slave...
String slaveReadString() {
  String s;
  uint8_t intRead;
  unsigned long prevTime = millis();

  // Read until newline character or timeout is reached
  prevTime = millis();
  while (millis() - prevTime < TEXT_TRANSMISSION_TIMEOUT) {
    if (Serial.available()) {
      intRead = Serial.read();
      // break the loop if end of transmission
      if (intRead == TRANSMISSION_END) break;
      s.concat((char)intRead);
    }
  }
  s.trim();
  // ToDo prefix with 'P0,0:' is positions are missing...
  if (
      !s.startsWith("LCDPRINT") && 
      !s.startsWith("CLR") && 
      !s.startsWith("BLINK") && 
      !s.startsWith("UNBLINK") && 
      !s.startsWith("VALUES") && 
      !s.startsWith("TONE") && 
      !s.startsWith("NOTONE")
    )
    s = "LCDPRINT0,0:"+s;
  return s;
}
  
void loop() {
  // Read (digital) push buttons...
  for (int i=0; i<7 ; i++) {
    if (digitalRead(buttons[i])<1) {
#ifdef DEBUG
      Serial.print("\nBUTTON_TRANSMISSION ");
      Serial.print(i+1);
      Serial.println(" TRANSMISSION_END");
#else
      Serial.write((uint8_t)BUTTON_TRANSMISSION);
      Serial.write((uint8_t)(i+1));
      Serial.write((uint8_t)TRANSMISSION_END);
#endif
      if (i==5 && !show_values) { // Manually toggle display: button 'E'
        lcd.clear();
        show_values = true;       
      }
    }
  }

  // Read and send joystick values as 'Steering'(X) and 'Throttle'(Y)...
  // ArduinoBlue: Throttle and Steering values go from 0 to 99
  current_x_value = map(analogRead(x_axis), 0, 1023, 0, 99);
  current_y_value = map(analogRead(y_axis), 0, 1023, 0, 99);
  if (current_x_value!=last_x_value || current_y_value!=last_y_value) {
    last_x_value = current_x_value;
    last_y_value = current_y_value;
#ifdef DEBUG
    Serial.print("\nlast_x_value=");
    Serial.print(last_x_value);
    Serial.print(" last_y_value=");
    Serial.println(last_y_value);
    Serial.print("DRIVE_TRANSMISSION ");
    Serial.print(current_y_value);
    Serial.print(' ');
    Serial.print(current_x_value);
    Serial.println(" TRANSMISSION_END");
#else
    Serial.write((uint8_t)DRIVE_TRANSMISSION);
    Serial.write((uint8_t)current_y_value);
    Serial.write((uint8_t)current_x_value);
    Serial.write((uint8_t)TRANSMISSION_END);
#endif
  }
  
  if (show_values) {
#ifndef DEBUG
  if (digitalRead(bt_state)==HIGH) {
    if (!bt_connected) {
      bt_connected = true;
    }
    // Write BTsymbol...
    lcd.setCursor(15,0);
    lcd.printByte(0);
  } else {
    if (bt_connected) {
      bt_connected = false;
    }
    // Remove BTsymbol...
    lcd.setCursor(15,0);
    lcd.print(" ");    
  }
#endif
    
    blink_line0 = false;
    blink_line1 = false;
        
    // Write values to LCD...
    lcd.setCursor(0,0);
    lcd.print(" X=     Y=    ");
    lcd.setCursor(3,0);
    lcd.print(current_x_value);
    lcd.setCursor(10,0);
    lcd.print(current_y_value);
    lcd.setCursor(1,1);
    lcd.print("Buttons");
    for (int i=0; i<7 ; i++) {
      lcd.setCursor(i+9,1);  
      if (digitalRead(buttons[i])==1) {
        lcd.printByte(0xA5);  
      } else
        lcd.printByte(0xDB);  
    }
  } else {
    if ((millis()-blinktime) > 500) { // Time to swap visibility...
      blinktime = millis();
      blink_show = !blink_show;
    }
    
    // Handle optional blinking per line...
    if (!blink_line0 && !showing_line0) {
        lcd.setCursor(0, 0);
        lcd.print(lcd_line0);
        showing_line0 = true;
    }
    if (blink_line0) {
      if (blink_show && !showing_line0) {
        lcd.setCursor(0, 0);
        lcd.print(lcd_line0);
        showing_line0 = true;
      }
      if (!blink_show && showing_line0) {
        lcd.setCursor(0, 0);
        for (int i=0; i<=MAX_COL; i++) lcd.print(" ");
        showing_line0 = false;
      }
    }

    // Handle optional blinking per line...
    if (!blink_line1 && !showing_line1) {
        lcd.setCursor(0, 1);
        lcd.print(lcd_line1);
        showing_line1 = true;
    } 
    if (blink_line1) {
      if (blink_show && !showing_line1) {
        lcd.setCursor(0, 1);
        lcd.print(lcd_line1);
        showing_line1 = true;
      }
      if (!blink_show && showing_line1) {
        lcd.setCursor(0, 1);
        for (int i=0; i<=MAX_COL; i++) lcd.print(" ");
        showing_line1 = false;
      }
    }
  }

  // Receiving actions...
  if (Serial.available() > 0) {
    // Receive and display text on a specific position...
    //   CLR[0|1] - Clears line 0 or 1, or both
    //   LCDPRINT{line},{col}:{message} - Text message 
    //      in which line = 0|1, col = 0~15 (for 1602 display) Example: P1,2:Sapperdeflap!
    //      To clear/overprint pad string with spaces, or use 'CLR' to clear whole lcd
    //   BLINK[0|1] - Blinks line 0 or 1, or both
    //   UNBLINK[0|1] - Unblinks line 0 or 1, or both
    //   VALUES - Returns to button/joystick monitoring
    //   TONE{freq}[duration] - Beeps piezo
    //   NOTONE - Silences piezo
    //
    uint8_t rx_value;
    receive_buffer = "";

#ifdef DEBUG
    receive_buffer = Serial.readString(); // Read from serial in...
    receive_buffer.trim();
#else
    rx_value = Serial.read(); // First read one char only...
    //if (rx_value == TRANSMISSION_END) break;
    if (rx_value==CONNECTION_CHECK)
        Serial.print(CONNECTION_CHECK);
    else if (rx_value==TEXT_SEND_TRANSMISSION)
        receive_buffer = slaveReadString();
#endif

    if (receive_buffer.startsWith("VALUES") && !show_values) {
      lcd.clear();
      show_values = true; // Start showing shields button/stick values...
    }
    
    if (receive_buffer.startsWith("CLR")) {
      if (receive_buffer.length()>3) {
        lcd_line = receive_buffer.substring(3).toInt()+0;
        if (lcd_line>MAX_LINE) lcd_line = MAX_LINE;
        if (lcd_line==0)
          lcd_line0 = empty_line;
        else
          lcd_line1 = empty_line;
        lcd.setCursor(0, lcd_line);
        lcd.print(empty_line);
      } else {
        lcd_line0 = empty_line;
        lcd_line1 = empty_line;
        lcd.clear();
      }
      
      show_values = false; // Stop showing shields button/stick values...
    }

    if (receive_buffer.startsWith("TONE")) {
      if (receive_buffer.length()>4) {
        tone_freq = tone_time = 0;
        tone_freq = receive_buffer.substring(4, receive_buffer.indexOf(',',0)).toInt();
        tone_time = receive_buffer.substring(receive_buffer.indexOf(',',0)+1).toInt();
        
#ifdef DEBUG
        Serial.print("freq=");
        Serial.print(tone_freq);
        Serial.print("duration=");
        Serial.print(tone_time);
#endif

        if (tone_freq>0)
          if (tone_time>0)
            tone(tone_pin, tone_freq, tone_time);
          else
            tone(tone_pin, tone_freq); // Keep tone...          
      } 
    }    

    if (receive_buffer.startsWith("NOTONE")) {
      if (receive_buffer.length()==6)
        noTone(tone_pin); // End tone...
    }    

    if (receive_buffer.startsWith("LCDPRINT")) {
      lcd_line = receive_buffer.substring(8, receive_buffer.indexOf(',',0)).toInt();
      lcd_col  = receive_buffer.substring(receive_buffer.indexOf(',',0)+1, receive_buffer.indexOf(':',0)).toInt();
      if (lcd_col>MAX_COL)   lcd_col = MAX_COL;
      if (lcd_line>MAX_LINE) lcd_line = MAX_LINE;
      if (show_values) lcd.clear();
      
      lcd.setCursor(lcd_col, lcd_line);
      
      if (lcd_line==0) {
        lcd_line0 = receive_buffer.substring(receive_buffer.indexOf(':',0)+1);
        lcd.print(lcd_line0);
      } else {
        lcd_line1 = receive_buffer.substring(receive_buffer.indexOf(':',0)+1);
        lcd.print(lcd_line1);
      }
      show_values = false; // Stop showing shields button/stick values...
    }

    if (receive_buffer.startsWith("BLINK")) {
      if (receive_buffer.length()>5) {
        lcd_line = receive_buffer.substring(5).toInt()+0;
        if (lcd_line>MAX_LINE) lcd_line = MAX_LINE;
        if (lcd_line==0) {
          blink_line0 = true;
        } else {
          blink_line1 = true;
        }
      } else {
        blink_line0 = true;
        blink_line1 = true;       
      }
    } 

    if (receive_buffer.startsWith("UNBLINK")) {
      if (receive_buffer.length()>7) {
        lcd_line = receive_buffer.substring(7).toInt()+0;
        if (lcd_line>MAX_LINE) lcd_line = MAX_LINE;
        if (lcd_line==0) {
          blink_line0 = false;
        } else {
          blink_line1 = false;
        }
      } else {
        blink_line0 = false;
        blink_line1 = false;  
      }
    }
  }

}
